//----------------------------------------------------------------------------------------------
// BSP : Seeed nRF52 Borads 1.1.1
// Board : Seeed nRF52 Borads / Seeed XIAO nRF52840 Sense 
//----------------------------------------------------------------------------------------------

// 2023/06/07
// Central sketch : nRF52_XIAO_LightSleepBatteryTest_central.ino
// Important : NVIC_SetPriority(RTC2_IRQn,3);

#include <Arduino.h>
#include <Adafruit_TinyUSB.h>   // For Serial.print()
#include <bluefruit.h>          // /Arduino15/packages/Seeeduino/hardware/nrf52/1.1.1/libraries/Bluefruit52Lib/src
#include <Adafruit_SPIFlash.h>  // Need to be deleted /Documents/Arduino/libraries/SdFat

#define HICHG         22        // Charge current　Hi:50mA Lo:100mA       D22 (13u)
#define CHG           23        // Charge LED Lo:ON                       D23 (17u)
#define VBAT_ENABLE   14        // Battery voltage read enable  Lo:enable D14 (14u)   
#define VBAT_READ     32        // Battery voltage read pin               D32 (31u)
#define SLEEP_TIME  3600        // Sleep time [sec]

#define dataNum 8               // Vbatt 4, mS 4 (2 data 8 byte)
union unionData {               // Union for data type conversion
  float     dataBuffFl[dataNum/4];
  uint32_t  dataBuff32[dataNum/4];
  uint8_t   dataBuff8[dataNum];
};
union unionData ud;
uint32_t mS, uS;
float Vbatt;                  // Battery voltage [V]
bool intFlag = true;          // Interrupu flag
bool connectedFlag = false;   // Connected flag

// Flash Class instance
Adafruit_FlashTransport_QSPI flashTransport;

// Custum Service and Characteristic
// 53c40000-8682-4dd1-be0c-40588193b485 for example
#define VbattService_UUID(val) (const uint8_t[]) { \
    0x85, 0xB4, 0x93, 0x81, 0x58, 0x40, 0x0C, 0xBE, 0xD1, 0x4D, \
    (uint8_t)(val & 0xff), (uint8_t)(val >> 8), \
    0x00, 0x00, 0xC4, 0x53 }

BLEService          VbattService         (VbattService_UUID(0x0000));
BLECharacteristic   dataCharacteristic   (VbattService_UUID(0x0030));

void setup()
{
  // Serial port initialization
  Serial.begin(115200);
//  while (!Serial);
  delay(2000);
  
  // Pin initialization
  pinMode(VBAT_READ, INPUT); 
  pinMode(HICHG, OUTPUT);
  pinMode(VBAT_ENABLE, OUTPUT);   
  pinMode(LED_RED, OUTPUT);
  pinMode(LED_GREEN, OUTPUT);
  pinMode(LED_BLUE, OUTPUT); 

  digitalWrite(HICHG, LOW);         // Charge current 100mA
  digitalWrite(VBAT_ENABLE, LOW);   // Battery voltage read enable   
  digitalWrite(LED_RED, HIGH);
  digitalWrite(LED_GREEN, HIGH);
  digitalWrite(LED_BLUE, HIGH);

  // ADC initialization　See wireing_analog_nRF52.c:73
  analogReference(AR_DEFAULT);      // Default 0.6V*6=3.6V
  analogReadResolution(12);         // wireing_analog_nRF52.c:39
  analogSampleTime(10);             // default 3uS　3,5,10,15,20,40

  // Enable DC-DC converter **************************************************************
  NRF_POWER->DCDCEN = 1;            // Enable DC/DC converter for REG1 stage

  // Flash power-down mode ***************************************************************
  flashTransport.begin();
  flashTransport.runCommand(0xB9);  // enter deep power-down mode
  delayMicroseconds(5);
  flashTransport.end();  

  // RTC initialization  
  initRTC(1024 * SLEEP_TIME);       // SLEEP_MTIME [sec]  

  // Initialization of Bruefruit class
  Bluefruit.configPrphBandwidth(BANDWIDTH_MAX);
  Bluefruit.configUuid128Count(15);

  Bluefruit.begin();
  Bluefruit.setTxPower(4);          // Check bluefruit.h for supported values
  Bluefruit.setConnLedInterval(50);
  Bluefruit.Periph.setConnectCallback(connect_callback);
  Bluefruit.Periph.setDisconnectCallback(disconnect_callback);

  // Custom Service Settings
  VbattService.begin();

  dataCharacteristic.setProperties(CHR_PROPS_NOTIFY | CHR_PROPS_READ);
  dataCharacteristic.setPermission(SECMODE_OPEN, SECMODE_NO_ACCESS);
  dataCharacteristic.setFixedLen(dataNum);
  dataCharacteristic.begin();

  // Advertisement Settings
  Bluefruit.Advertising.addFlags(BLE_GAP_ADV_FLAGS_LE_ONLY_GENERAL_DISC_MODE);
  Bluefruit.Advertising.addTxPower();
  Bluefruit.Advertising.addService(VbattService);
  Bluefruit.ScanResponse.addName();
  
  Bluefruit.Advertising.restartOnDisconnect(true);
  Bluefruit.Advertising.setIntervalMS(20, 153);     // fast mode 20mS, slow mode 153mS
  Bluefruit.Advertising.setFastTimeout(30);         // fast mode 30 sec
  Bluefruit.Advertising.start(0);                   // 0 = Don't stop advertising after n seconds

  Serial.println("End of initialization");
}

//*************************************************************************************************************
void loop()
{
  if(intFlag == true)     // If wake up
  {       
    Serial.println("Connecting to Central ....");
    while(!Bluefruit.connected()) {
    }
    Serial.println("Connected to Central");  
    
    if (Bluefruit.connected())      // If connected to the central
    {
      digitalWrite(LED_GREEN, LOW);

      // Battery voltage averaging
      int Vadc = 0;
      for (int i = 0; i < 16; i++) {
        Vadc = Vadc + analogRead(VBAT_READ);
      }
      Vadc = Vadc / 16;
      Vbatt = 2.961 * 3.6 * Vadc / 4096;   // Vref=3.6  ratio=2.961
      ud.dataBuffFl[0] = Vbatt;

      // Measurement time
      mS = millis();
      ud.dataBuff32[1] = mS;

      Serial.print(Vbatt, 3);
      Serial.print(" ");
      Serial.print(mS);
      Serial.println(); 

      // Data transfer
      dataCharacteristic.notify(ud.dataBuff8, dataNum);

      delayMicroseconds(100000);      // 100mS for LED
      digitalWrite(LED_GREEN, HIGH);
    }    
    intFlag = false;
    connectedFlag = false;        
  }  
  // System ON Sleep ******************************************************************
  __WFI();
  __SEV();
  __WFI(); 
}

//**************************************************************************************************************
// callback function
void connect_callback(uint16_t conn_handle)
{
  // Get the reference to current connection
  BLEConnection* connection = Bluefruit.Connection(conn_handle);

  char central_name[32] = { 0 };
  connection->getPeerName(central_name, sizeof(central_name));

  Serial.print("【connect_callback】 Connected to ");
  Serial.println(central_name);

  connectedFlag = true;
}

void disconnect_callback(uint16_t conn_handle, uint8_t reason)
{
  (void) conn_handle;
  (void) reason;

  Serial.print("【disconnect_callback】 Disconnected, reason = 0x");
  Serial.println(reason, HEX);

  connectedFlag = false;
}

//**************************************************************************************************************
// RTC initialization
void initRTC(unsigned long RTCcount) 
{
  // See datasheet "6.22 RTC - Real-time counter 6.22.10 Registers"
  NRF_CLOCK->TASKS_LFCLKSTOP = 1;
  NRF_CLOCK->LFCLKSRC = 1;       // select LFXO
  NRF_CLOCK->TASKS_LFCLKSTART = 1;
  while(NRF_CLOCK->LFCLKSTAT != 0x10001)
  
  NRF_RTC2->TASKS_STOP = 1;      // stop counter 
  NRF_RTC2->TASKS_CLEAR = 1;     // clear counter
  NRF_RTC2->PRESCALER = 31;      // set counter prescaler, fout=32768/(1+PRESCALER)　1024Hz
  NRF_RTC2->CC[0] = RTCcount;    // set value for TRC compare register 0
  NRF_RTC2->INTENSET = 0x10000;  // enable interrupt CC[0] compare match event
  NRF_RTC2->EVTENCLR = 0x10000;  // clear counter when CC[0] event
  NVIC_SetPriority(RTC2_IRQn,3); // Higher priority than SoftDeviceAPI
  NVIC_EnableIRQ(RTC2_IRQn);     // enable interrupt  
  NRF_RTC2->TASKS_START = 1;     // start Timer
}

//**************************************************************************************************************
// RTC interrupt handler
extern "C" void RTC2_IRQHandler(void)
{
  if ((NRF_RTC2->EVENTS_COMPARE[0] != 0) && ((NRF_RTC2->INTENSET & 0x10000) != 0)) {
    NRF_RTC2->EVENTS_COMPARE[0] = 0;  // clear compare register 0 event
    NRF_RTC2->TASKS_CLEAR = 1;        // clear counter

    intFlag = true;   
    digitalWrite(LED_RED, LOW);
    delayMicroseconds(10000);         // 10mS for LED
    digitalWrite(LED_RED, HIGH);   
  }
}
